﻿using ICSV.Interface.Types;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using XnaGeometry;

namespace ICSV.Server
{
    /// <summary>
    /// This class is responsible for finding axes for the given images.
    /// It will rotate the whole coordinate system if necessary to preserve consistency in appearence.
    /// </summary>
    public class AxesBuilder
    {
        /// <summary>
        /// Creates axes for the given images.
        /// This method will update the given coordinates in the 'items' parameter if necessary.
        /// </summary>
        /// <remarks>
        /// This method tries to rotate the 'L' axis in such a way that Lmin (black) is at Y- and Lmax (white) is at Y+.
        /// The 'a' axis will be rotated such that Amin (green) is pointing to X- and Amax (red) is pointing to X+.
        /// The 'b' axis will be flipped if necessary, such that Bmin (blue) is pointing towards Z+ and Bmax (yellow) is pointing to Z-.
        /// Please note, that the resulting axes don't need to be orthogonal, since the MDS algorithm does not ensure an orthogonal system.
        /// </remarks>
        /// <param name="axisImages">The images of the axes. Useful for coloring the axes.</param>
        /// <param name="axisCoords">The coordinates of the axes. Necessary for rotating the whole system.</param>
        /// <param name="items">The images in the current coordinate system, with their corresponding coordinates.
        /// The coordinates will be adapted when rotating/flipping the coordinate system.</param>
        /// <param name="dimensions">The dimension of the coordinate system. Either 2D or 3D.</param>
        /// <returns>the resulting axes.</returns>
        public AxisUpdate[] CreateAxes(ImageContainer[] axisImages, CoordinateContainer[] axisCoords, Dictionary<ImageContainer, CoordinateContainer> items, int dimensions)
        {
            // AXIS ORDER: Lmin, Lmax, Amin, Amax, Bmin, Bmax
            var Lmin = axisImages[0]; var Lmax = axisImages[1];
            var Amin = axisImages[2]; var Amax = axisImages[3];
            var Bmin = axisImages[4]; var Bmax = axisImages[5];

            Vector4 LminCoord = axisCoords[0].GetCoord(dimensions);
            Vector4 LmaxCoord = axisCoords[1].GetCoord(dimensions);
            Vector4 AminCoord = axisCoords[2].GetCoord(dimensions);
            Vector4 AmaxCoord = axisCoords[3].GetCoord(dimensions);
            Vector4 BminCoord = axisCoords[4].GetCoord(dimensions);
            Vector4 BmaxCoord = axisCoords[5].GetCoord(dimensions);

            // align axes center points
            // reference center point is the center point of 'L' axis

            var lcx = (LminCoord.X + LmaxCoord.X) / 2;
            var lcy = (LminCoord.Y + LmaxCoord.Y) / 2;
            var lcz = (LminCoord.Z + LmaxCoord.Z) / 2;

            var acx = (AminCoord.X + AmaxCoord.X) / 2;
            var acy = (AminCoord.Y + AmaxCoord.Y) / 2;
            var acz = (AminCoord.Z + AmaxCoord.Z) / 2;

            var bcx = (BminCoord.X + BmaxCoord.X) / 2;
            var bcy = (BminCoord.Y + BmaxCoord.Y) / 2;
            var bcz = (BminCoord.Z + BmaxCoord.Z) / 2;
            
            AminCoord.X -= (acx - lcx); AminCoord.Y -= (acy - lcy); AminCoord.Z -= (acz - lcz);
            AmaxCoord.X -= (acx - lcx); AmaxCoord.Y -= (acy - lcy); AmaxCoord.Z -= (acz - lcz);
            BminCoord.X -= (bcx - lcx); BminCoord.Y -= (bcy - lcy); BminCoord.Z -= (bcz - lcz);
            BmaxCoord.X -= (bcx - lcx); BmaxCoord.Y -= (bcy - lcy); BmaxCoord.Z -= (bcz - lcz);

            // rotate the axes and the content
            // such that Lmax is going up (Y+) and Amin is going to the left (X-) and Bmin is pointing to the front (Z+)

            // matrix format is m_{column}{row}
            // create translation matrices to move the axes from/to origin
            Matrix to = new Matrix(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, -lcx, -lcy, -lcz, 1);

            // we want to rotate the L axis:
            Vector4 dir = LmaxCoord - LminCoord;
            Vector3 dir3d = new Vector3(dir.X, dir.Y, dir.Z);
            dir3d.Normalize();

            // with quaternions - match L axis with Y axis
            var Q = QuaternionFrom2Vectors(dir3d, new Vector3(0, 1, 0));
            Matrix rotationL = new Matrix();
            Quaternion.ToMatrix(Q, out rotationL);

            Matrix M = to * rotationL; // 1. translation, 2. rotation

            // bring to origin and rotate L axis:
            LminCoord = Vector4.Transform(LminCoord, M);
            LmaxCoord = Vector4.Transform(LmaxCoord, M);
            AminCoord = Vector4.Transform(AminCoord, M);
            AmaxCoord = Vector4.Transform(AmaxCoord, M);
            BminCoord = Vector4.Transform(BminCoord, M);
            BmaxCoord = Vector4.Transform(BmaxCoord, M);

            // rotate Amin around Y, such that it is left of the L axis
            dir3d = new Vector3(AminCoord.X, 0, AminCoord.Z); // project onto Y = 0
            dir3d.Normalize();

            double angleToX = Math.Acos(Vector3.Dot(dir3d, new Vector3(-1, 0, 0)));
            double cosAngle = Math.Cos(angleToX);
            double sinAngle = Math.Sign(AminCoord.Z) * Math.Sin(angleToX);

            Matrix rotY = new Matrix(cosAngle, 0, sinAngle, 0, 0, 1, 0, 0, -sinAngle, 0, cosAngle, 0, 0, 0, 0, 1);

            // rotate A axis:
            Matrix M2 = rotY;
            LminCoord = Vector4.Transform(LminCoord, M2);
            LmaxCoord = Vector4.Transform(LmaxCoord, M2);
            AminCoord = Vector4.Transform(AminCoord, M2);
            AmaxCoord = Vector4.Transform(AmaxCoord, M2);
            BminCoord = Vector4.Transform(BminCoord, M2);
            BmaxCoord = Vector4.Transform(BmaxCoord, M2);

            // mirror B axis (if necessary):
            Matrix M3 = Matrix.Identity;
            if (BminCoord.Z < 0 && dimensions == 3) // don't do that with 2 dimensions
            {
                // mirror all coordinates around Z = 0 plane (i.e. all Z-coordinates should be inverted)
                M3 = new Matrix(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1);
            }

            // flip B axis:
            LminCoord = Vector4.Transform(LminCoord, M3);
            LmaxCoord = Vector4.Transform(LmaxCoord, M3);
            AminCoord = Vector4.Transform(AminCoord, M3);
            AmaxCoord = Vector4.Transform(AmaxCoord, M3);
            BminCoord = Vector4.Transform(BminCoord, M3);
            BmaxCoord = Vector4.Transform(BmaxCoord, M3);

            // adjust length of axes:
            int length = 300;
            LminCoord.Normalize(); LmaxCoord.Normalize();
            AminCoord.Normalize(); AmaxCoord.Normalize();
            BminCoord.Normalize(); BmaxCoord.Normalize();

            LminCoord = Vector4.Multiply(LminCoord, length);
            LmaxCoord = Vector4.Multiply(LmaxCoord, length);
            AminCoord = Vector4.Multiply(AminCoord, length);
            AmaxCoord = Vector4.Multiply(AmaxCoord, length);
            BminCoord = Vector4.Multiply(BminCoord, length);
            BmaxCoord = Vector4.Multiply(BmaxCoord, length);

            // transform all images according to new (rotated) coordinate system:
            Matrix Mglobal = M * M2 * M3;
            foreach (var entry in items)
            {
                var imageCoords = entry.Value;
                Vector4 coords = imageCoords.GetCoord(dimensions);
                coords = Vector4.Transform(coords, Mglobal);
                imageCoords.SetCoord(dimensions, coords);
            }

            // publish updated axes:
            AxisUpdate[] newAxes = new AxisUpdate[3];
            newAxes[0] = CreateAxisUpdate(Lmin, Lmax, LminCoord, LmaxCoord, dimensions);
            newAxes[1] = CreateAxisUpdate(Amin, Amax, AminCoord, AmaxCoord, dimensions);
            newAxes[2] = CreateAxisUpdate(Bmin, Bmax, BminCoord, BmaxCoord, dimensions);

            return newAxes;
        }

        private static AxisUpdate CreateAxisUpdate(ImageContainer dataP1, ImageContainer dataP2, Vector4 coordP1, Vector4 coordP2, int dimensions)
        {
            return new AxisUpdate
            {
                XCoordP1 = coordP1.X,
                YCoordP1 = coordP1.Y,
                ZCoordP1 = coordP1.Z,
                XCoordP2 = coordP2.X,
                YCoordP2 = coordP2.Y,
                ZCoordP2 = coordP2.Z,
                ColorP1 = ConvertColorToInt(dataP1.FirstRGBcolor),
                ColorP2 = ConvertColorToInt(dataP2.FirstRGBcolor)
            };
        }

        private static int ConvertColorToInt(Color color)
        {
            return (color.R << 16) + (color.G << 8) + color.B;
        }

        /* Build a unit quaternion representing the rotation
         * from u to v. The input vectors need to be normalized.
         * http://lolengine.net/blog/2014/02/24/quaternion-from-two-vectors-final
         */
        private static Quaternion QuaternionFrom2Vectors(Vector3 u, Vector3 v)
        {
            double dotProduct = Vector3.Dot(u, v);

            Vector3 axis;
            if ((dotProduct + 1) < 0.000001)
            {
                /* If u and v are exactly opposite, rotate 180 degrees
                 * around an arbitrary orthogonal axis. Axis normalisation
                 * can happen later, when we normalise the quaternion. */
                axis = Math.Abs(u.X) > Math.Abs(u.Z) ? new Vector3(-u.Y, u.X, 0) : new Vector3(0, -u.Z, u.Y);
            }
            else
            {
                /* Otherwise, build quaternion the standard way. */
                axis = Vector3.Cross(u, v);
            }

            double angle = Math.Acos(dotProduct);
            axis.Normalize();

            Quaternion q = Quaternion.CreateFromAxisAngle(axis, -angle);
            q.Normalize();
            return q;
        }
    }
}
