﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using XnaGeometry;

namespace ICSV.Server
{
    /// <summary>
    /// Callback interface. Useful for status updates of the <see cref="CoordinateProvider"/>.
    /// </summary>
    public interface CoordinateCallback
    {
        /// <summary>
        /// Called as soon as a bunch coordinates for images have been updated by the provider.
        /// </summary>
        /// <param name="items">a dictionary with the image as key and its corresponding coordinates</param>
        void coordinatesUpdated(Dictionary<ImageContainer, CoordinateContainer> items);

        /// <summary>
        /// The provider has finished. No more updates will be sent after this method is called.
        /// </summary>
        void finished();

        /// <summary>
        /// An error occurred.
        /// </summary>
        /// <param name="message">the error message</param>
        void erroneous(String message);

        /// <summary>
        /// Status information by the provider
        /// </summary>
        /// <param name="message">the information message</param>
        void info(String message);
    }

    /// <summary>
    /// Data container class, holding 2D and 3D coordinates.
    /// </summary>
    public class CoordinateContainer
    {
        private Vector4 Coord2D { get; set; } // Z = 0, W = 1
        private Vector4 Coord3D { get; set; } // W = 1

        /// <summary>
        /// Get coordinates for the given dimensions (either 2D or 3D).
        /// </summary>
        /// <param name="dimensions">2 or 3</param>
        /// <returns>a vector with 'w' component 1. 'z' = 0 in case of 2D.</returns>
        public Vector4 GetCoord(int dimensions)
        {
            if (dimensions == 2) return Coord2D;
            if (dimensions == 3) return Coord3D;
            throw new ArgumentException("Only 2D/3D coordinates available");
        }

        /// <summary>
        /// Setting the coordinates for 2D or 3D.
        /// </summary>
        /// <param name="dimensions">2 or 3</param>
        /// <param name="coord">the vector to set</param>
        public void SetCoord(int dimensions, Vector4 coord)
        {
            if (dimensions == 2) Coord2D = coord;
            else if (dimensions == 3) Coord3D = coord;
            else throw new ArgumentException("Only 2D/3D coordinates available");
        }
    }

    /// <summary>
    /// This class is responsible for calculating coordinates for images.
    /// </summary>
    /// <remarks>
    /// This class is chained after the <see cref="DistanceProvider"/> and received both images and distances in between them.
    /// With the distances a MDS algorithm is started (for 2D and 3D Euclidean space) and the coordinates according to the distances
    /// are calculated. Finally the coordinates are scaled to fit in a specific region.
    /// </remarks>
    public class CoordinateProvider : DistanceCallback
    {
        // wait for at least this count until we calculate new coordinates
        // higher count -> lesser updates on client side, but less computation
        private static readonly int MINIMUM_OBJECTS = 20;

        private static readonly double IMAGE_DELIVERY_FACTOR = 0.75; // smaller is faster (0.65 is very fast)

        private readonly CoordinateCallback _callback;
        private DistanceProvider _provider;

        private bool _distanceProviderFinished = false;
        private int _updated = 0;

        private Object _lockObject = new Object();

        private TaskFactory taskFactory;
        private CancellationToken _cancelToken;

        /// <summary>
        /// Creates a new instance, using the callback to communicate progress.
        /// </summary>
        /// <param name="callback">the callback will be notified in case of status updates</param>
        /// <param name="cancelToken">the cancellation token is used to cancel the current progress.</param>
        public CoordinateProvider(CoordinateCallback callback, CancellationToken cancelToken)
        {
            _callback = callback;
            _cancelToken = cancelToken;

            // Create a scheduler that uses only one thread
            LimitedConcurrencyLevelTaskScheduler lcts = new LimitedConcurrencyLevelTaskScheduler(1);
            taskFactory = new TaskFactory(lcts);
        }

        /// <summary>
        /// Sets the distance provider, necessary to get distances for images.
        /// </summary>
        /// <param name="distanceProv">the provider</param>
        public void setDistanceProvider(DistanceProvider distanceProv)
        {
            _provider = distanceProv;
        }

        private int calculateMinimumObjectCountToStart()
        {
            int currentImageCount = _provider.GetImageCount();
            int tmp = (int)Math.Round(Math.Pow(currentImageCount, IMAGE_DELIVERY_FACTOR));
            return Math.Max(tmp, MINIMUM_OBJECTS);
        }

        void tryStartProcessingInternal()
        {
            bool shouldStart;
            bool wasFinished;
            int minObjCount = calculateMinimumObjectCountToStart();
            lock (_lockObject)
            {
                wasFinished = _distanceProviderFinished;
                shouldStart = _updated > 0 && (_updated >= minObjCount || _distanceProviderFinished);
                if (shouldStart)
                {
                    _updated = 0;
                    _distanceProviderFinished = false;
                }
            }
            if (!shouldStart)
            {
                if (wasFinished)
                {
                    _callback.finished();
                }
                return;
            }

            if (_cancelToken.IsCancellationRequested) return;

            ImageContainer[] images;
            double[][] distanceMatrix = _provider.GetDistanceMatrix(out images);
            if (_cancelToken.IsCancellationRequested) return;

            int count = images.Length;

            Stopwatch sw = new Stopwatch();
            sw.Start();

            // Multi Dimensional Scaling
            int stresschange = 1; // stress change between iterations should be <= 10^(-stresschange)
            int maxIterations = 0; // maximum iterations, 0 for no bounds
            string report2D;
            double[][] coord2D = ICSV.Server.MDS.MDS.Mds(out report2D, distanceMatrix, 2, maxIterations, 0, stresschange); // result: 2 x N matrix
            if (_cancelToken.IsCancellationRequested) return;

            string report3D;
            double[][] coord3D = ICSV.Server.MDS.MDS.Mds(out report3D, distanceMatrix, 3, maxIterations, 0, stresschange); // result: 3 x N matrix
            if (_cancelToken.IsCancellationRequested) return;

            sw.Stop();
            int imageCount = count - 6; // minus 6 images (= 3 axes with 2 images)
            _callback.info("MDS 2D report: " + report2D);
            _callback.info("MDS 3D report: " + report3D);
            _callback.info("Multidimensional Scaling (MDS) for " + imageCount + " images done in " + Math.Round(sw.ElapsedMilliseconds / 1000.0, 3) + " seconds");

            // convert to Vector4 positions and check for errors:
            double maxP2, minP2, minY2;
            Vector4[] pos2D = checkValidAndComputeBounds(coord2D, 2, out minP2, out maxP2, out minY2);

            double maxP3, minP3, minY3;
            Vector4[] pos3D = checkValidAndComputeBounds(coord3D, 3, out minP3, out maxP3, out minY3);

            if (pos2D == null || pos3D == null)
            {
                _callback.erroneous("MDS failed. Possibly a non-metric space!");
                return;
            }

            // scale the positions, such that they are not too far spread apart:
            scalePositions(pos2D, 2, minP2, maxP2, minY2);
            scalePositions(pos3D, 3, minP3, maxP3, minY3);

            // wrap the results up and notify listener
            Dictionary<ImageContainer, CoordinateContainer> result = new Dictionary<ImageContainer, CoordinateContainer>();
            for (int i = 0; i < count; i++)
            {
                CoordinateContainer container = new CoordinateContainer();

                container.SetCoord(2, pos2D[i]);
                container.SetCoord(3, pos3D[i]);
                result.Add(images[i], container);
            }

            _callback.coordinatesUpdated(result);
            if (wasFinished)
            {
                _callback.finished();
            }
        }

        void tryStartProcessing()
        {
            // switch thread:
            taskFactory.StartNew(() =>
            {
                tryStartProcessingInternal();
            }, _cancelToken);
        }

        Vector4[] checkValidAndComputeBounds(double[][] coordinates, int dimensions, out double minP, out double maxP, out double minY)
        {
            maxP = Double.MinValue;
            minP = Double.MaxValue;
            minY = Double.MaxValue;

            int count = coordinates[0].Length;

            Vector4[] positions = new Vector4[count];

            for (int j = 0; j < count; j++)
            {
                if (Double.IsNaN(coordinates[0][j]) || Double.IsNaN(coordinates[1][j]))
                {
                    // THIS SHOULD NOT HAPPEN ANYMORE

                    // typically this is because the MDS algorithm has been used on a non-metric space!
                    // The EMD is a metric distance if:
                    // - the ground distance is metric
                    // - the sum of weights of each signature must be equal
                    //      - this means that all images must have the same pixel count
                    //      - and there must not be any threshold used when creating the signature

                    // at this point it is typically to late to do anything (except for another, non-metric MDS algorithm)
                    // so just log the event and end calculations
                    Console.WriteLine("MDS failed. Possible non-metric space!");
                    return null;
                }
                minP = Math.Min(minP, coordinates[0][j]);
                maxP = Math.Max(maxP, coordinates[0][j]);

                minP = Math.Min(minP, coordinates[1][j]);
                maxP = Math.Max(maxP, coordinates[1][j]);
                minY = Math.Min(minY, coordinates[1][j]);

                if (dimensions == 3)
                {
                    minP = Math.Min(minP, coordinates[2][j]);
                    maxP = Math.Max(maxP, coordinates[2][j]);
                    positions[j] = new Vector4(coordinates[0][j], coordinates[1][j], coordinates[2][j], 1);
                }
                else // dimensions == 2
                {
                    positions[j] = new Vector4(coordinates[0][j], coordinates[1][j], 0, 1);
                }
            }

            return positions;
        }

        void scalePositions(Vector4[] positions, int dimensions, double minP, double maxP, double minY)
        {
            double scaleMin = -200;
            double scaleMax = 200;

            for (int i = 0; i < positions.Length; i++)
            {
                var p = positions[i];

                p.X = (p.X - minP) / (maxP - minP) * (scaleMax - scaleMin) + scaleMin;
                p.Y = (p.Y - minP) / (maxP - minP) * (scaleMax - scaleMin) + scaleMin;
                p.Y -= (minY - minP) / (maxP - minP) * (scaleMax - scaleMin) + scaleMin - 20; // move above zero plane

                if (dimensions == 3)
                {
                    p.Z = (p.Z - minP) / (maxP - minP) * (scaleMax - scaleMin) + scaleMin;
                }

                positions[i] = p;
            }
        }

        public void distancesUpdated(ImageContainer image)
        {
            lock (_lockObject)
            {
                _updated++;
            }
            tryStartProcessing();
        }

        public void finished()
        {
            lock (_lockObject)
            {
                _distanceProviderFinished = true;
            }
            tryStartProcessing();
        }

        public void erroneous(String message)
        {
            _callback.erroneous(message);
        }

        public void info(String message)
        {
            _callback.info(message);
        }
    }
}
