﻿using ICSV.Server.Signatures;
using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Diagnostics;
using System.Linq;
using System.Runtime.Caching;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace ICSV.Server
{
    /// <summary>
    /// The callback interface used by the provider to report status updates.
    /// </summary>
    public interface DistanceCallback
    {
        /// <summary>
        /// Distances for the given image have been calculated.
        /// As soon as this method is called distances from/to this image to all other images, currently held by
        /// the distance provider will be available.
        /// </summary>
        /// <param name="image"></param>
        void distancesUpdated(ImageContainer image);

        /// <summary>
        /// The provider has finished. No more updates will occur after this method has been called.
        /// </summary>
        void finished();

        /// <summary>
        /// An error has occurred.
        /// </summary>
        /// <param name="message">the error message</param>
        void erroneous(String message);

        /// <summary>
        /// Status information by the provider
        /// </summary>
        /// <param name="message">the information message</param>
        void info(String message);
    }

    /// <summary>
    /// Internal class for identifying a pair of images.
    /// The pair is uniquely identified by a combination of the identifiers of both images.
    /// </summary>
    class ImagePair
    {
        private string id1, id2;
        public string IdCombined { get; private set; }

        /// <summary>
        /// A new image pair with the given images
        /// </summary>
        /// <param name="i1">image 1</param>
        /// <param name="i2">image 2</param>
        public ImagePair(ImageContainer i1, ImageContainer i2)
        {
            id1 = i1.Identifier;
            id2 = i2.Identifier;
            if (id1.CompareTo(id2) < 0)
            {
                IdCombined = id1 + id2;
            }
            else
            {
                IdCombined = id2 + id1;
            }
        }

        /// <summary>
        /// Whether or not this object is equal to the given one.
        /// </summary>
        /// <param name="obj">The object to test equality against.</param>
        /// <returns>true if equal, otherwise false</returns>
        public override bool Equals(System.Object obj)
        {
            if (obj == null) return false;

            // If parameter cannot be cast to ImagePair return false.
            ImagePair p = obj as ImagePair;
            if ((System.Object)p == null) return false;

            // Return true if pairs match
            return IdCombined.Equals(p.IdCombined);
        }

        /// <summary>
        /// Whether or not this object is equal to the given one.
        /// </summary>
        /// <param name="p">The object to test equality against.</param>
        /// <returns>true if equal, otherwise false</returns>
        public bool Equals(ImagePair p)
        {
            // If parameter is null return false:
            if ((object)p == null) return false;

            // Return true if pairs match
            return IdCombined.Equals(p.IdCombined);
        }

        /// <summary>
        /// The hash code of this image pair
        /// </summary>
        /// <returns>the hash code</returns>
        public override int GetHashCode()
        {
            return IdCombined.GetHashCode();
        }
    }

    /// <summary>
    /// This provider is responsible for calculating Earth Mover Distances between images.
    /// It gets notified about new images by the <see cref="ImageProvider"/>.
    /// It works asynchronously and uses multiple threads to effectively calculate the EMDs.
    /// </summary>
    public class DistanceProvider : ImageProviderCallback
    {
        private readonly DistanceCallback _callback;
        private bool _imageProviderFinished = false;
        private int _processing = 0;

        private TaskFactory taskFactory;
        private CancellationToken _cancelToken;

        private Object _lockObject = new Object();
        private List<ImageContainer> _images = new List<ImageContainer>();
        private List<ImageContainer> _receivedImages = new List<ImageContainer>();
        private Dictionary<ImagePair, double> _distances = new Dictionary<ImagePair, double>();

        /// <summary>
        /// A new instance, using the callback for status updates and the cancellation token in case it gets cancelled.
        /// </summary>
        /// <param name="callback">the callback for continous status updates</param>
        /// <param name="cancelToken">the cancellation token, controlling the possibility to cancel the provider's work</param>
        public DistanceProvider(DistanceCallback callback, CancellationToken cancelToken)
        {
            _callback = callback;
            _cancelToken = cancelToken;

            // Create a scheduler with a fixed pool of threads:
            LimitedConcurrencyLevelTaskScheduler lcts = new LimitedConcurrencyLevelTaskScheduler(8);
            taskFactory = new TaskFactory(lcts);
        }

        /// <summary>
        /// Returns the current image count this provider has calculated distances for.
        /// </summary>
        /// <returns></returns>
        public int GetImageCount()
        {
            lock (_lockObject)
            {
                return _images.Count;
            }
        }

        /// <summary>
        /// Returns an 'adjacency' matrix, containing the distances between already calculated images.
        /// All images with finished distance calculations are returned in the 'images' out-parameter.
        /// </summary>
        /// <param name="images">will contain all images processed so far</param>
        /// <returns>the distance matrix</returns>
        public double[][] GetDistanceMatrix(out ImageContainer[] images)
        {
            lock (_lockObject)
            {
                images = _images.ToArray();
            }

            int count = images.Length;

            double[][] distanceMatrix = new double[count][];
            for (int i = 0; i < count; i++)
            {
                distanceMatrix[i] = new double[count];
            }

            for (int i = 0; i < count; i++)
            {
                for (int j = i + 1; j < count; j++)
                {
                    double distance = getDistance(images[i], images[j]);
                    distanceMatrix[j][i] = distanceMatrix[i][j] = distance;
                }
            }

            return distanceMatrix;
        }

        private double getDistance(ImageContainer image1, ImageContainer image2)
        {
            ImagePair pair = new ImagePair(image1, image2);
            lock (_lockObject)
            {
                if (!_distances.ContainsKey(pair)) return 0;
                return _distances[pair];
            }
        }

        private void addImageInternal(ImageContainer image, bool suppressEvents)
        {
            // Calculate the distance of this image to all other images saved
            int idx = 0;
            while (true)
            {
                if (_cancelToken.IsCancellationRequested)
                {
                    suppressEvents = true;
                    break;
                }

                ImageContainer otherImage;
                lock (_lockObject)
                {
                    if (idx == _images.Count)
                    {
                        // we reached the end of the list
                        // -> add current image and we are done
                        _images.Add(image);
                        break;
                    }
                    otherImage = _images[idx];
                    idx++;
                }

                // calculate the distance for the image pair:
                ImagePair imagePair = new ImagePair(image, otherImage);
                double distance = calculateDistance(imagePair, image.Signature, otherImage.Signature);
                lock (_lockObject)
                {
                    _distances.Add(imagePair, distance);
                }
            }

            if (!suppressEvents)
            {
                _callback.distancesUpdated(image);
            }
        }

        private double calculateDistance(ImagePair imagePair, Signature sig1, Signature sig2)
        {
            string key = imagePair.IdCombined;
            // 3/4 of the maximum distance in CIE Lab color space between two points (calculated with Matlab)
            // Note:
            // It does not really matter which mass penalty is used with FastEMD
            // maximum distance, maximum distance * 0.75 and -1 pretty much ends up in the same result.
            // Don't use a mass penalty of 0, because otherwise we would need a non-metric MDS
            double maximumDistance = 258.6826988242628 * 0.75;
            Lazy<double> loader = new Lazy<double>(() => FastEMD.FastEMD.distance(sig1, sig2, maximumDistance));

            var lazyObject = MemoryCache.Default.AddOrGetExisting(key, loader, new CacheItemPolicy()) as Lazy<double>;
            return (lazyObject ?? loader).Value;
        } 

        /// <summary>
        /// Manually add an image to calculate distance for.
        /// </summary>
        /// <param name="image">the image to add</param>
        /// <param name="suppressEvents">If true, then no calls to the callback will be made.</param>
        public void addImage(ImageContainer image, bool suppressEvents)
        {
            lock (_lockObject)
            {
                if (_receivedImages.Contains(image))
                {
                    // could happen if Flickr provides the same image link twice
                    // -> avoid double computation
                    return;
                }
                _receivedImages.Add(image);
                _processing++;
            }

            // switch thread:
            Task t = taskFactory.StartNew(() =>
            {
                addImageInternal(image, suppressEvents);
            }, _cancelToken);


            t.ContinueWith((something) =>
            {
                lock (_lockObject)
                {
                    _processing--;
                    checkFinished();
                }
            });
        }

        public void foundImage(ImageContainer image)
        {
            addImage(image, false);
        }

        private void checkFinished()
        {
            bool finished;
            lock (_lockObject)
            {
                finished = (_imageProviderFinished && _processing == 0);
                if (finished)
                {
                    _imageProviderFinished = false; // reset
                }
            }
            if (finished)
            {
                _callback.finished();
            }
        }

        public void finished()
        {
            lock (_lockObject)
            {
                _imageProviderFinished = true;
            }
            checkFinished();
        }

        public void erroneous(String message)
        {
            _callback.erroneous(message);
        }

        public void info(String message)
        {
            _callback.info(message);
        }
    }
}
