﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using ICSV.Server.Signatures;

namespace ICSV.Server.FastEMD
{
    /// <summary>
    /// See https://github.com/telmomenezes/JFastEMD/blob/master/src/com/telmomenezes/jfastemd/JFastEMD.java
    /// </summary>
    /// <remarks> 
    /// This class computes the Earth Mover's Distance, using the EMD-HAT algorithm
    /// created by Ofir Pele and Michael Werman.
    /// 
    /// This implementation is strongly based on the C++ code by the same authors,
    /// that can be found here:
    /// http://www.cs.huji.ac.il/~ofirpele/FastEMD/code/
    /// 
    /// Some of the author's comments on the original were kept or edited for 
    /// this context.
    /// 
    /// @author Telmo Menezes (telmo@telmomenezes.com)
    /// </remarks>
    class FastEMD
    {
        /// <summary>
        /// This interface is similar to Rubner's interface. See:
        /// http://www.cs.duke.edu/~tomasi/software/emd.htm
        /// 
        /// This method computes the EMD between the given two signatures.
        /// </summary>
        /// <remarks>
        /// To get the same results as Rubner's code you should set extra_mass_penalty to 0,
        /// and divide by the minimum of the sum of the two signature's weights. However, I
        /// suggest not to do this as you lose the metric property and more importantly, in my
        /// experience the performance is better with emd_hat. for more on the difference
        /// between emd and emd_hat, see the paper:
        /// A Linear Time Histogram Metric for Improved SIFT Matching
        /// Ofir Pele, Michael Werman
        /// ECCV 2008
        ///
        /// To get shorter running time, set the ground distance function to
        /// be a thresholded distance. For example: min(L2, T). Where T is some threshold.
        /// Note that the running time is shorter with smaller T values. Note also that
        /// thresholding the distance will probably increase accuracy. Finally, a thresholded
        /// metric is also a metric. See paper:
        /// Fast and Robust Earth Mover's Distances
        /// Ofir Pele, Michael Werman
        /// ICCV 2009
        ///
        /// If you use this code, please cite the papers.
        /// </remarks>
        /// <param name="signature1">The first signature</param>
        /// <param name="signature2">The second signature</param>
        /// <param name="extraMassPenalty">The penalty for extra mass.
        /// If you want the resulting distance to be a metric, it should be
        /// at least half the diameter of the space (maximum possible distance between any two points). If you
        /// want partial matching you can set it to zero (but then the resulting distance is not guaranteed to be a metric).
        /// Default value is -1 which means the mass penalty is set to the maximum ground-distance between the two signatures.
        /// See also the 'emdHat' function.
        /// </param>
        /// <returns>the calculated EMD</returns>
        static public double distance(Signature signature1, Signature signature2, double extraMassPenalty)
        {
            List<double> P = new List<double>();
            List<double> Q = new List<double>();
            for (int i = 0; i < signature1.getNumberOfFeatures() + signature2.getNumberOfFeatures(); i++)
            {
                P.Add(0.0);
                Q.Add(0.0);
            }
            for (int i = 0; i < signature1.getNumberOfFeatures(); i++)
            {
                P[i] = signature1.getWeights()[i];
            }
            for (int j = 0; j < signature2.getNumberOfFeatures(); j++)
            {
                Q[j + signature1.getNumberOfFeatures()] = signature2.getWeights()[j];
            }

            List<List<double>> C = new List<List<double>>();
            for (int i = 0; i < P.Count; i++)
            {
                List<double> vec = new List<double>();
                for (int j = 0; j < P.Count; j++)
                {
                    vec.Add(0.0);
                }
                C.Add(vec);
            }
            for (int i = 0; i < signature1.getNumberOfFeatures(); i++)
            {
                for (int j = 0; j < signature2.getNumberOfFeatures(); j++)
                {
                    double dist = signature1.getFeatures()[i]
                            .groundDist(signature2.getFeatures()[j]);
                    if (dist < 0) throw new InvalidOperationException();
                    C[i][j + signature1.getNumberOfFeatures()] = dist;
                    C[j + signature1.getNumberOfFeatures()][i] = dist;
                }
            }

            return emdHat(P, Q, C, extraMassPenalty);
        }


        static private long emdHatImplLongLongInt(List<long> Pc, List<long> Qc,
                List<List<long>> C, long extraMassPenalty) {

        int N = Pc.Count;
        if (Qc.Count != N) throw new ArgumentException();

        // Ensuring that the supplier - P, have more mass.
        // Note that we assume here that C is symmetric
        List<long> P;
        List<long> Q;
        long absDiffSumPSumQ;
        long sumP = 0;
        long sumQ = 0;
        for (int i = 0; i < N; i++)
            sumP += Pc[i];
        for (int i = 0; i < N; i++)
            sumQ += Qc[i];
        if (sumQ > sumP) {
            P = Qc;
            Q = Pc;
            absDiffSumPSumQ = sumQ - sumP;
        } else {
            P = Pc;
            Q = Qc;
            absDiffSumPSumQ = sumP - sumQ;
        }

        // creating the b List that contains all vertexes
        List<long> b = new List<long>();
        for (int i = 0; i < 2 * N + 2; i++) {
            b.Add(0);
        }
        int THRESHOLD_NODE = 2 * N;
        int ARTIFICIAL_NODE = 2 * N + 1; // need to be last !
        for (int i = 0; i < N; i++) {
            b[i] = P[i];
        }
        for (int i = N; i < 2 * N; i++) {
            b[i] = Q[i - N];
        }

        // remark*) I put here a deficit of the extra mass, as mass that flows
        // to the threshold node
        // can be absorbed from all sources with cost zero (this is in reverse
        // order from the paper,
        // where incoming edges to the threshold node had the cost of the
        // threshold and outgoing
        // edges had the cost of zero)
        // This also makes sum of b zero.
        b[THRESHOLD_NODE] = -absDiffSumPSumQ;
        b[ARTIFICIAL_NODE] = 0;

        long maxC = 0;
        for (int i = 0; i < N; i++) {
            for (int j = 0; j < N; j++) {
                if (C[i][j] < 0) throw new InvalidOperationException();
                if (C[i][j] > maxC)
                    maxC = C[i][j];
            }
        }
        if (extraMassPenalty == -1)
            extraMassPenalty = maxC;

        ISet<int> sourcesThatFlowNotOnlyToThresh = new HashSet<int>();
        ISet<int> sinksThatGetFlowNotOnlyFromThresh = new HashSet<int>();
        long preFlowCost = 0;

        // regular edges between sinks and sources without threshold edges
        List<List<Edge>> c = new List<List<Edge>>();
        for (int i = 0; i < b.Count; i++) {
            c.Add(new List<Edge>());
        }
        for (int i = 0; i < N; i++) {
            if (b[i] == 0)
                continue;
            for (int j = 0; j < N; j++) {
                if (b[j + N] == 0)
                    continue;
                if (C[i][j] == maxC)
                    continue;
                c[i].Add(new Edge(j + N, C[i][j]));
            }
        }

        // checking which are not isolated
        for (int i = 0; i < N; i++) {
            if (b[i] == 0)
                continue;
            for (int j = 0; j < N; j++) {
                if (b[j + N] == 0)
                    continue;
                if (C[i][j] == maxC)
                    continue;
                sourcesThatFlowNotOnlyToThresh.Add(i);
                sinksThatGetFlowNotOnlyFromThresh.Add(j + N);
            }
        }

        // converting all sinks to negative
        for (int i = N; i < 2 * N; i++) {
            b[i] = -b[i];
        }

        // add edges from/to threshold node,
        // note that costs are reversed to the paper (see also remark* above)
        // It is important that it will be this way because of remark* above.
        for (int i = 0; i < N; ++i) {
            c[i].Add(new Edge(THRESHOLD_NODE, 0));
        }
        for (int j = 0; j < N; ++j) {
            c[THRESHOLD_NODE].Add(new Edge(j + N, maxC));
        }

        // artificial arcs - Note the restriction that only one edge i,j is
        // artificial so I ignore it...
        for (int i = 0; i < ARTIFICIAL_NODE; i++) {
            c[i].Add(new Edge(ARTIFICIAL_NODE, maxC + 1));
            c[ARTIFICIAL_NODE].Add(new Edge(i, maxC + 1));
        }

        // remove nodes with supply demand of 0
        // and vertexes that are connected only to the
        // threshold vertex
        int currentNodeName = 0;
        // Note here it should be List<int> and not List<int>
        // as I'm using -1 as a special flag !!!
        int REMOVE_NODE_FLAG = -1;
        List<int> nodesNewNames = new List<int>();
        List<int> nodesOldNames = new List<int>();
        for (int i = 0; i < b.Count; i++) {
            nodesNewNames.Add(REMOVE_NODE_FLAG);
            nodesOldNames.Add(0);
        }
        for (int i = 0; i < N * 2; i++) {
            if (b[i] != 0) {
                if (sourcesThatFlowNotOnlyToThresh.Contains(i)
                        || sinksThatGetFlowNotOnlyFromThresh.Contains(i)) {
                    nodesNewNames[i] = currentNodeName;
                    nodesOldNames.Add(i);
                    currentNodeName++;
                } else {
                    if (i >= N) {
                        preFlowCost -= (b[i] * maxC);
                    }
                    b[THRESHOLD_NODE] = b[THRESHOLD_NODE] + b[i]; // add mass(i<N) or deficit (i>=N)
                }
            }
        }
        nodesNewNames[THRESHOLD_NODE] = currentNodeName;
        nodesOldNames.Add(THRESHOLD_NODE);
        currentNodeName++;
        nodesNewNames[ARTIFICIAL_NODE] = currentNodeName;
        nodesOldNames.Add(ARTIFICIAL_NODE);
        currentNodeName++;

        List<long> bb = new List<long>();
        for (int i = 0; i < currentNodeName; i++) {
            bb.Add(0);
        }

        int jj = 0;
        for (int i = 0; i < b.Count; i++) {
            if (nodesNewNames[i] != REMOVE_NODE_FLAG) {
                bb[jj] = b[i];
                jj++;
            }
        }

        List<List<Edge>> cc = new List<List<Edge>>();
        for (int i = 0; i < bb.Count; i++) {
            cc.Add(new List<Edge>());
        }
        for (int i = 0; i < c.Count; i++) {
            if (nodesNewNames[i] == REMOVE_NODE_FLAG)
                continue;
            foreach (Edge it in c[i]) {
                if (nodesNewNames[it._to] != REMOVE_NODE_FLAG) {
                    cc[nodesNewNames[i]].Add(
                            new Edge(nodesNewNames[it._to], it._cost));
                }
            }
        }

        MinCostFlow mcf = new MinCostFlow();

        long myDist;

        List<List<Edge0>> flows = new List<List<Edge0>>(bb.Count);
        for (int i = 0; i < bb.Count; i++) {
            flows.Add(new List<Edge0>());
        }

        long mcfDist = mcf.compute(bb, cc, flows);

        myDist = preFlowCost + // pre-flowing on cases where it was possible
                mcfDist + // solution of the transportation problem
                (absDiffSumPSumQ * extraMassPenalty); // emd-hat extra mass penalty

        return myDist;
    }

        static private double emdHat(List<double> P, List<double> Q, List<List<double>> C,
                double extraMassPenalty)
        {

            // This condition should hold:
            // ( 2^(sizeof(CONVERT_TO_T*8)) >= ( MULT_FACTOR^2 )
            // Note that it can be problematic to check it because
            // of overflow problems. I simply checked it with Linux calc
            // which has arbitrary precision.
            double MULT_FACTOR = 1000000;

            // Constructing the input
            int N = P.Count;
            List<long> iP = new List<long>();
            List<long> iQ = new List<long>();
            List<List<long>> iC = new List<List<long>>();
            for (int i = 0; i < N; i++)
            {
                iP.Add(0);
                iQ.Add(0);
                List<long> vec = new List<long>();
                for (int j = 0; j < N; j++)
                {
                    vec.Add(0);
                }
                iC.Add(vec);
            }

            // Converting to CONVERT_TO_T
            double sumP = 0.0;
            double sumQ = 0.0;
            double maxC = C[0][0];
            for (int i = 0; i < N; i++)
            {
                sumP += P[i];
                sumQ += Q[i];
                for (int j = 0; j < N; j++)
                {
                    if (C[i][j] > maxC)
                        maxC = C[i][j];
                }
            }
            double minSum = Math.Min(sumP, sumQ);
            double maxSum = Math.Max(sumP, sumQ);
            double PQnormFactor = MULT_FACTOR / maxSum;
            double CnormFactor = MULT_FACTOR / maxC;
            for (int i = 0; i < N; i++)
            {
                iP[i] = (long)(Math.Floor(P[i] * PQnormFactor + 0.5));
                iQ[i] = (long)(Math.Floor(Q[i] * PQnormFactor + 0.5));
                for (int j = 0; j < N; j++)
                {
                    if (maxC == 0)
                    {
                        iC[i][j] = 0;
                    }
                    else
                    {
                        iC[i][j] = (long)(Math.Floor(C[i][j] * CnormFactor + 0.5));
                    }
                }
            }

            // computing distance without extra mass penalty
            double dist = emdHatImplLongLongInt(iP, iQ, iC, 0);
            // unnormalize
            dist = dist / PQnormFactor;
            dist = dist / CnormFactor;

            // adding extra mass penalty
            if (extraMassPenalty == -1)
                extraMassPenalty = maxC;
            dist += (maxSum - minSum) * extraMassPenalty;

            return dist;
        }
    }
}
