﻿using ICSV.Server.Images;
using ICSV.Server.Signatures;
using FlickrNet;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Runtime.Caching;
using System.Collections.Specialized;
using System.Diagnostics;

namespace ICSV.Server
{
    /// <summary>
    /// Interface for the callback registered at the <see cref="ImageProvider"/>.
    /// The callback is valid for one search session.
    /// <seealso cref="ImageProvider.searchImages"/>
    /// </summary>
    public interface ImageProviderCallback
    {
        /// <summary>
        /// Will be called as soon as an image has been found.
        /// The given 'image' is already initialized with all relevant data (e.g. signature, data, color, ...)
        /// </summary>
        /// <param name="image">the image which has been found (e.g. on Flickr) and has been initialized</param>
        void foundImage(ImageContainer image);

        /// <summary>
        /// A previous search has been finished.
        /// </summary>
        void finished();

        /// <summary>
        /// An error occurred during searching.
        /// </summary>
        /// <param name="message">The error message</param>
        void erroneous(String message);

        /// <summary>
        /// A status update for the current ongoing search.
        /// </summary>
        /// <param name="message">The info message</param>
        void info(String message);
    }

    /// <summary>
    /// This class holds all relevant information about an image, before distances are calculated and a position is applied.
    /// </summary>
    public class ImageContainer
    {
        /// <summary>
        /// A unique identifier for each image.
        /// </summary>
        public readonly string Identifier;

        /// <summary>
        /// The signature of the image. It contains color clusters with weights (i.e. the count of pixels in the cluster).
        /// </summary>
        public readonly Signature Signature;

        /// <summary>
        /// The data URI for the image in the format 'data:image/...'
        /// </summary>
        /// <remarks>
        /// Data URI scheme: http://en.wikipedia.org/wiki/Data_URI_scheme
        /// </remarks>
        public readonly string DataURI;

        /// <summary>
        /// The RGB color of the first pixel of the image.
        /// This is useful for the axes of the color space, which are initially represented by images, consisting only of one color.
        /// </summary>
        public readonly Color FirstRGBcolor;

        /// <summary>
        /// To construct an object all parameters are considered to be mandatory.
        /// </summary>
        /// <param name="id">the unique id</param>
        /// <param name="sig">the calculated signature</param>
        /// <param name="uri">the data URI</param>
        /// <param name="firstColor">the color of the first pixel</param>
        public ImageContainer(string id, Signature sig, string uri, Color firstColor)
        {
            Identifier = id;
            Signature = sig;
            DataURI = uri;
            FirstRGBcolor = firstColor;
        }

        /// <summary>
        /// Whether or not this object is equal to the given object.
        /// Two objects are equal if the ids are equal.
        /// </summary>
        /// <param name="obj">the object to check equality with</param>
        /// <returns>true if equal, false otherwise</returns>
        public override bool Equals(System.Object obj)
        {
            if (obj == null) return false;

            ImageContainer p = obj as ImageContainer;
            if ((System.Object)p == null) return false;

            // Return true if id matches
            return Identifier.Equals(p.Identifier);
        }

        /// <summary>
        /// Whether or not this object is equal to the given object.
        /// Two objects are equal if the ids are equal.
        /// </summary>
        /// <param name="obj">the object to check equality with</param>
        /// <returns>true if equal, false otherwise</returns>
        public bool Equals(ImageContainer p)
        {
            if ((object)p == null) return false;
            return Identifier.Equals(p.Identifier);
        }

        /// <summary>
        /// The hashcode of the object. I.e. the hashcode of the identifier.
        /// </summary>
        /// <returns>the hashcode</returns>
        public override int GetHashCode()
        {
            return Identifier.GetHashCode();
        }
    }

    /// <summary>
    /// This class is responsible for searching images on the web (i.e. on Flickr) or loading local images.
    /// After loading, signatures are created, which can be then used to calculate the Earth Mover's Distance.
    /// </summary>
    /// <remarks>
    /// This class uses a cache to allow fast access to previously loaded images. Images are cached according to their unique ids.
    /// This class is a singleton. The only instance is available through ImageProvider.Instance.
    /// </remarks>
    public class ImageProvider
    {
        // all images are resized to this dimensions:
        // smaller means faster computation (e.g. convertion to Lab color space), but it is less precise
        private static readonly int IMAGE_LARGE_EDGE = 100;
        private static readonly int IMAGE_SMALL_EDGE = 75;

        private static readonly int MAX_THREADS;
        
        /// <summary>
        /// The singleton instance of this class.
        /// </summary>
        public static readonly ImageProvider Instance;

        // static constructor
        static ImageProvider()
        {
            MAX_THREADS = 8; // more than 8 is problematic with the Flickr API
            Instance = new ImageProvider();
        }

        // ------------------

        private Object lockObject = new Object();
        private AutoResetEvent autoEvent = new AutoResetEvent(false);

        private HashSet<string> currentlySearching = new HashSet<string>();

        private TaskFactory taskFactory;

        private ImageProvider()
        {
            // Create a scheduler that uses a limited pool of threads
            LimitedConcurrencyLevelTaskScheduler lcts = new LimitedConcurrencyLevelTaskScheduler(MAX_THREADS);
            taskFactory = new TaskFactory(lcts);
        }

        /// <summary>
        /// Search Flickr with the given options and maximum count.
        /// A callback needs to be provided for progress and status report.
        /// This method works asynchronously.
        /// </summary>
        /// <param name="flickr">the Flickr API object</param>
        /// <param name="options">contains all relevant settings for searching. E.g. sort order</param>
        /// <param name="photoCount">the maximum number of photos to retrieve</param>
        /// <param name="callback">the callback for status updates</param>
        /// <param name="cancelToken">the cancellation token, in case the search needs to be cancelled</param>
        public void searchImages(Flickr flickr, PhotoSearchOptions options, int photoCount, ImageProviderCallback callback, CancellationToken cancelToken)
        {
            List<Task> tasks = new List<Task>();
            int photosLeft = photoCount;
            int currentPage = 1;
            options.PerPage = Math.Min(photoCount, 500); // a maximum of 500 per page is allowed according to Flickr API
           
            while (photosLeft > 0)
            {
                options.Page = currentPage;
                PhotoCollection photos;
                try
                {
                    photos = flickr.PhotosSearch(options);
                }
                catch (WebException e)
                {
                    callback.erroneous(e.Message);
                    return;
                }
                catch (FlickrException e)
                {
                    callback.erroneous(e.Message);
                    return;
                }

                // --- process photos:
                foreach (var p in photos)
                {
                    string imageURL = p.Small320Url;
                    Task t = taskFactory.StartNew(() =>
                    {
                        WebClient webClient = new WebClient(); // WARNING! one WebClient, one download at a time
                        ImageContainer container = getRemoteImage(imageURL, webClient, callback);
                        callback.foundImage(container);
                    }, cancelToken);
                    tasks.Add(t);
                    photosLeft--;
                    if (photosLeft == 0) break;
                }
                currentPage++;
                if (photos.Count < options.PerPage)
                {
                    // we don't have more pictures than that
                    callback.info("Flickr does not provide more images for the given search query");
                    break;
                }
            }

            try
            {
                Task.WaitAll(tasks.ToArray());
            }
            catch (AggregateException)
            {
                // ignore. we were probably cancelled
            }
            callback.finished();
        }

        private ImageContainer getCachedImageOrStartLoading(string url, Lazy<ImageContainer> loader)
        {
            string key = url;
            var lazyObject = MemoryCache.Default.AddOrGetExisting(key, loader, new CacheItemPolicy()) as Lazy<ImageContainer>;
            return (lazyObject ?? loader).Value;
            // operator ??: if left operand is not null use it, otherwise use the right operand
        }

        private ImageContainer getRemoteImage(string remoteURL, WebClient webClient, ImageProviderCallback callback)
        {
            Lazy<ImageContainer> loader = new Lazy<ImageContainer>(() => loadRemoteImage(remoteURL, webClient, callback));
            return getCachedImageOrStartLoading(remoteURL, loader);
        }

        /// <summary>
        /// Loads an image from the given remote URL, using the web client and providing status reports to the callback.
        /// </summary>
        /// <param name="remoteURL">from where to load</param>
        /// <param name="webClient">the client to use for loading</param>
        /// <param name="callback">the callback for status updates</param>
        /// <returns>the loaded image with all necessary fields initialized</returns>
        public ImageContainer loadRemoteImage(string remoteURL, WebClient webClient, ImageProviderCallback callback)
        {
            callback.info("Attempting to load " + remoteURL);

            //Get byte representation
            byte[] imageBytes = webClient.DownloadData(remoteURL);

            ImageContainer container;
            using (MemoryStream ms = new MemoryStream(imageBytes))
            {
                using (Image img = Image.FromStream(ms))
                {
                    using (Image newImg = minimizeImage(img))
                    {
                        container = createImageContainer(newImg, imageBytes, remoteURL);
                    }
                }
            }
            return container;
        }

        /// <summary>
        /// Synchronously loads a local image from the given local server url.
        /// </summary>
        /// <param name="localURL">the URL to load from</param>
        /// <returns>the fully initialized image container</returns>
        public ImageContainer getLocalImage(string localURL)
        {
            Lazy<ImageContainer> loader = new Lazy<ImageContainer>(() => loadLocalImage(localURL));
            return getCachedImageOrStartLoading(localURL, loader);
        }

        private ImageContainer loadLocalImage(string localURL)
        {
            // --- load image data:
            using (Image img = Image.FromFile(localURL))
            {
                using (Image newImg = minimizeImage(img))
                {
                    byte[] byteData = (byte[]) new ImageConverter().ConvertTo(newImg, typeof(byte[]));
                    return createImageContainer(newImg, byteData, localURL);
                }
            }
        }

        private Image minimizeImage(Image image)
        {
            // one side of the image is constant (e.g. 320) the other side is random
            // we want to resize the images to largeEdge x smallEdge so that all images have the same pixel count and
            // therefore the EMD is a true metric:
            int newWidth = IMAGE_LARGE_EDGE;
            int newHeight = IMAGE_SMALL_EDGE;
            if (image.Height > image.Width)
            {
                // portrait mode
                newHeight = IMAGE_LARGE_EDGE;
                newWidth = IMAGE_SMALL_EDGE;
            }

            return ResizeImage(image, newWidth, newHeight);
        }

        private ImageContainer createImageContainer(Image image, byte[] byteData, string URL)
        {
            Bitmap bmp = new Bitmap(image);
            var firstColor = bmp.GetPixel(0, 0);

            // --- convert to CIE Lab color space:
            var dataURI = GetBase64DataURI(byteData, URL);
            double[][] labData = ColorSpaceConverter.ConvertRgbToLab(bmp);

            // --- create signature:
            //double minCellSize = 12.5; // more features per signature -> EMD will be slower with calculation
            double minCellSize = 25; // about 6x faster than 12.5, good performance and good quality (recommended in paper 'The Earth Mover's Distance as a Metric for Image Retrieval')
            double threshold = 0.001; // remove clusters with insignificant weights (less than 0.1 % of pixel count)
            var signature = new SignatureProvider().CreateSignature(labData, minCellSize, threshold);

            // --- create result container:
            ImageContainer container = new ImageContainer(URL, signature, dataURI, firstColor);
            return container;
        }

        /// <summary>
        /// Resize the image to the specified width and height.
        /// Based on http://stackoverflow.com/questions/1922040/resize-an-image-c-sharp
        /// </summary>
        /// <param name="image">The image to resize.</param>
        /// <param name="width">The width to resize to.</param>
        /// <param name="height">The height to resize to.</param>
        /// <returns>The resized image.</returns>
        private static Bitmap ResizeImage(Image image, int width, int height)
        {
            var destRect = new System.Drawing.Rectangle(0, 0, width, height);
            var destImage = new Bitmap(width, height);

            destImage.SetResolution(image.HorizontalResolution, image.VerticalResolution);

            using (var graphics = Graphics.FromImage(destImage))
            {
                graphics.CompositingMode = System.Drawing.Drawing2D.CompositingMode.SourceCopy;
                graphics.CompositingQuality = System.Drawing.Drawing2D.CompositingQuality.HighSpeed;
                graphics.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
                graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.Default;
                graphics.PixelOffsetMode = System.Drawing.Drawing2D.PixelOffsetMode.HighSpeed;

                using (var wrapMode = new ImageAttributes())
                {
                    wrapMode.SetWrapMode(System.Drawing.Drawing2D.WrapMode.TileFlipXY);
                    graphics.DrawImage(image, destRect, 0, 0, image.Width, image.Height, GraphicsUnit.Pixel, wrapMode);
                }
            }

            return destImage;
        }

        private static string GetBase64DataURI(byte[] imageBytes, string originalUrl)
        {
            // Data URI scheme: http://en.wikipedia.org/wiki/Data_URI_scheme
            System.Uri uri = new Uri(originalUrl);
            string format = System.IO.Path.GetExtension(uri.AbsolutePath).Replace(".", ""); // e.g. jpg

            var b64String = System.Convert.ToBase64String(imageBytes);
            return "data:image/" + format + ";base64," + b64String;
        }

    }
}
