﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;

namespace ICSV.Server.Images
{
    /// <summary>
    /// Converts <see cref="Bitmap"/> data from sRGB -> XYZ -> CIE Lab color space, using D65 white point.
    /// </summary>
    public class ColorSpaceConverter
    {
        /// <summary>
        /// Matrix for RGB to XYZ conversion
        /// </summary>
        private static readonly double[,] RGB_2_XYZ = new double[3, 3];

        /// <summary>
        /// D65 white point definition
        /// </summary>
        private static readonly double[] D65_XnYnZn = {0.95047, 1, 1.0883};

        static ColorSpaceConverter() // static constructor
        {
            // see http://www.brucelindbloom.com/Eqn_RGB_XYZ_Matrix.html
            // sRGB (D65 white point)
            RGB_2_XYZ[0, 0] = 0.4124564;
            RGB_2_XYZ[0, 1] = 0.3575761;
            RGB_2_XYZ[0, 2] = 0.1804375;
            RGB_2_XYZ[1, 0] = 0.2126729;
            RGB_2_XYZ[1, 1] = 0.7151522;
            RGB_2_XYZ[1, 2] = 0.0721750;
            RGB_2_XYZ[2, 0] = 0.0193339;
            RGB_2_XYZ[2, 1] = 0.1191920;
            RGB_2_XYZ[2, 2] = 0.9503041;
        }

        private ColorSpaceConverter()
        {
            // private constructor -> everything's static
        }

        /// <summary>
        /// Converts the given sRGB bitmap to a double array. Resulting in values in CIE Lab color space.
        /// </summary>
        /// <param name="bitmap">
        /// The bitmap can be 8, 24 or 32 (with alpha) bit.
        /// </param>
        /// <returns>
        /// An array of size bitmap.width * bitmap.height, where each entry has 3 elements (L,a,b)
        /// </returns>
        public static double[][] ConvertRgbToLab(Bitmap bitmap)
        {
            LockBitmap lb = new LockBitmap(bitmap);
            lb.LockBits();

            int skip = 1; // Skip every 'skip-1' pixel (for performance)

            int size = lb.Width * lb.Height / skip;

            //jagged array, NOT 2D array! eg labPoint[i] returns array with [l,a,b] values
            double[][] labPoints = new double[size][];

            int step = lb.Step;
            int depth = lb.Depth;
            byte[] data = lb.Pixels;

            if (depth == 32 || depth == 24) // RGBA, RGB
            {
                int nr = 0;
                for (int i = 0; i < data.Length; i += step * skip)
                {
                    byte b = data[i];
                    byte g = data[i + 1];
                    byte r = data[i + 2];
                    double[] pointInXYZ = ConvertRgbToXYZ(r, g, b);
                    labPoints[nr] = ConvertXyzToLab(pointInXYZ);
                    nr++;
                }
            }
            else if (depth == 8)
            // For 8 bpp get color value (Red, Green and Blue values are the same)
            {
                int nr = 0;
                for (int i = 0; i < data.Length; i += step * skip)
                {
                    byte c = data[i];
                    double[] pointInXYZ = ConvertRgbToXYZ(c, c, c);
                    labPoints[nr] = ConvertXyzToLab(pointInXYZ);
                    nr++;
                }
            }

            lb.UnlockBits(false);
            return labPoints; 
        }

        /// <summary>
        /// Converts the given RGB values in the range of [0,255] to XYZ coordinates.
        /// </summary>
        /// <param name="r">red value</param>
        /// <param name="g">green value</param>
        /// <param name="b">blue value</param>
        /// <returns>
        /// An array with [0] = X, [1] = Y, [2] = Z.
        /// X,Y,Z are roughly in the range of [0, 1]
        /// </returns>
        public static double[] ConvertRgbToXYZ(byte r, byte g, byte b)
        {
            double red = r / 255.0;
            double green = g / 255.0;
            double blue = b / 255.0;

            double rl = LinearizeRGB(red);
            double gl = LinearizeRGB(green);
            double bl = LinearizeRGB(blue);

            double[] dArray = new double[3];
            dArray[0] = RGB_2_XYZ[0, 0] * rl + RGB_2_XYZ[0, 1] * gl + RGB_2_XYZ[0, 2] * bl;
            dArray[1] = RGB_2_XYZ[1, 0] * rl + RGB_2_XYZ[1, 1] * gl + RGB_2_XYZ[1, 2] * bl;
            dArray[2] = RGB_2_XYZ[2, 0] * rl + RGB_2_XYZ[2, 1] * gl + RGB_2_XYZ[2, 2] * bl;

            return dArray;
        }

        /// <summary>
        /// Converts the XYZ coordinates of <see cref="ConvertRgbToXYZ"/> to CIE Lab color space coordinates.
        /// </summary>
        /// <param name="vector">
        /// [0] = X, [1] = Y, [2] = Z
        /// </param>
        /// <returns>
        /// An array with [0] = L, [1] = a, [2] = b.
        /// L (lightness) is in the range of [0,100].
        /// a (green and red/magenta opposing channel) is in the range of roughly [-87, 99]
        /// b (blue and yellow opposing channel) is in the range of roughly [-108, 95]
        /// </returns>
        public static double[] ConvertXyzToLab(double[] vector)
        {
            // Convert to L*a*b* using standard Illuminant D65 XnYnZn = (0.95047, 1, 1.0883)

            // divide by white point:
            double xr = vector[0] / D65_XnYnZn[0];
            double yr = vector[1] / D65_XnYnZn[1];
            double zr = vector[2] / D65_XnYnZn[2];

            double lx = ToLab(xr);
            double ly = ToLab(yr);
            double lz = ToLab(zr);
        
            double l = 116 * ly - 16;
            double a = 500 * (lx - ly);
            double b = 200 * (ly - lz);

            double [] labArray = new double[3];
            labArray[0] = l;
            labArray[1] = a;
            labArray[2] = b;

            return labArray;
        }

        /// <summary>
        /// Linearize RGB channels.
        /// See http://www.ryanjuckett.com/programming/rgb-color-space-conversion/
        /// </summary>
        /// <param name="sRGBColor">Normalized R/G/B color in range [0,1]</param>
        /// <returns>The linearized color</returns>
        private static double LinearizeRGB(double sRGBColor)
        {
            if (sRGBColor <= 0.04045)
                return sRGBColor / 12.92;
            else
                return Math.Pow(((sRGBColor + 0.055) / 1.055), 2.4);
        }

        /// <summary>
        /// Helper method for conversion to Lab color space.
        /// See http://en.wikipedia.org/wiki/Lab_color_space#CIELAB-CIEXYZ_conversions
        /// </summary>
        /// <param name="coord">The X/Y/Z coordinate, already normalized according to white point</param>
        /// <returns>The f(t) helper function of the Wikipedia article.</returns>
        private static double ToLab(double coord)
        {
            if (coord > 0.008856) // (6/29)^3
                return Math.Pow(coord, (1.0 / 3));
            else
                return (841 * coord) / 108 + 4 / 29; // 1/3 * (29/2)^2 * coord + 4/29
        }
    }
}

 