﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using ServiceStack.ServiceHost;

using FlickrNet;
using System.Net;

using ICSV.Interface.Operations;
using ICSV.Interface;
using ICSV.Interface.Types;
using System.Threading; 

namespace ICSV.Server
{
	/// <summary>
	/// Web-Service (via ServiceStack framework) to control the server.
	/// Responsible for starting a new image search, cancelling, retrieving status, ...
	/// </summary>
	public class IntroService
		: IService
	{
		private readonly EMDConfig config;

		private static System.Object lockObject = new System.Object();
		private static int currentSearchId = 0;
		private static readonly Dictionary<int, PrepareImages> imageTasks = new Dictionary<int, PrepareImages>();

		/// <summary>
		/// New instance with the given config.
		/// Called by the ServiceStack framework!
		/// </summary>
		/// <param name="config"></param>
		public IntroService(EMDConfig config)
		{
			this.config = config;
		}

		/// <summary>
		/// Web request to cancel an ongoing search.
		/// </summary>
		/// <param name="request">the request with additional information</param>
		public void Any(CancelSearch request)
		{
			var searchId = request.ForSearchId;

			PrepareImages prep;
			lock (lockObject)
			{
				if (!imageTasks.ContainsKey(searchId)) throw new InvalidOperationException("invalid search id");
				prep = imageTasks[searchId];
			}

			prep.Cancel();
		}

		/// <summary>
		/// Web request to get current status objects from the server.
		/// </summary>
		/// <param name="request">the request with additional information</param>
		/// <returns>a <see cref="GetStatusResponse"/> object, containing one or more status updates.</returns>
		public object Any(GetStatus request)
		{
			var searchId = request.ForSearchId;

			PrepareImages prep;
			lock (lockObject)
			{
				if (!imageTasks.ContainsKey(searchId)) throw new InvalidOperationException("invalid search id");
				prep = imageTasks[searchId];
			}

			var status = prep.GetStatus(1);
			return new GetStatusResponse() { SearchId = searchId, Status = new Interface.Types.ArrayOfStatusData(status) };
		}

		/// <summary>
		/// Web request to start a new image search.
		/// </summary>
		/// <param name="request">the request with additional information</param>
		/// <returns>A <see cref="SearchImagesResponse"/> object with a new search id.</returns>
		public object Any(SearchImages request)
		{
			PrepareImages prep = new PrepareImages(request, config);
			int searchId;
			lock (lockObject)
			{
				currentSearchId++;
				searchId = currentSearchId;
				imageTasks.Add(searchId, prep);
			}
			prep.Start();

			return new SearchImagesResponse { SearchId = searchId }; 
		}

		/// <summary>
		/// Web request to get image results for a currently ongoing or finished search.
		/// </summary>
		/// <param name="request">the request with additional information</param>
		/// <returns>A <see cref="GetImageResultsResponse"/> object.</returns>
		public object Any(GetImageResults request)
		{
			var searchId = request.ForSearchId;

			PrepareImages prep;
			lock (lockObject)
			{
				if (!imageTasks.ContainsKey(searchId)) throw new InvalidOperationException("invalid search id");
				prep = imageTasks[searchId];
			}

			bool finished;
			AxisUpdate[] axes2D, axes3D;
			ImageData[] images;
			prep.GetResults(out finished, out images, out axes2D, out axes3D);

			if (finished)
			{
				// remove the image task in 2 seconds:
				new Timer(RemoveImageTask, searchId, 2000, System.Threading.Timeout.Infinite);
			}

			return new GetImageResultsResponse() {
				IsFinalResult = finished,
				SearchId = searchId,
				Images = new Interface.Types.ArrayOfImageData(images),
				Axes2D = new Interface.Types.ArrayOfAxisUpdate(axes2D),
				Axes3D = new Interface.Types.ArrayOfAxisUpdate(axes3D)
			};
		}

		/// <summary>
		/// Will be called asynchronously to remove an image task from the list of current tasks.
		/// This will cleanup everything the image task holds.
		/// </summary>
		/// <param name="state">the search id of the image task</param>
		public void RemoveImageTask(object state)
		{
			int searchId = (int) state;
			lock (lockObject)
			{
				if (imageTasks.ContainsKey(searchId))
				{
					imageTasks.Remove(searchId);
				}
			}
		}
	}
}
