﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using MDS;

namespace ICSV.Server.MDS
{
    /// <summary>
    /// Easy access interface to the MDSJ component for Multi-Dimensional Scaling
    /// </summary>
    /// <remarks>
    /// MDSJ available at http://www.inf.uni-konstanz.de/algo/software/mdsj/
    /// Documentation at http://www.inf.uni-konstanz.de/algo/software/mdsj/javadoc/
    /// </remarks>
    public class MDS
    {
        /// <summary>
        /// Runs a MDS algorithm.
        /// More precisely an approximation using landmark points.
        /// <seealso cref="V. de Silva, J.B. Tenenbaum, Sparse multidimensional scaling using landmark points. "/>
        /// </summary>
        /// <param name="report">a documentation of the result's precision and remaining stress</param>
        /// <param name="D">the square-sized distance matrix used to compute positions</param>
        /// <param name="dimensions">the dimension of the resulting Euclidean space. Typically 2 or 3.</param>
        /// <param name="rounds">the maximum count of iterations used for computation. If the parameter is 0 it is ignored.</param>
        /// <param name="timeout">the maximum time allowed for computation. In milliseconds. If the parameter is 0 it is ignored.</param>
        /// <param name="stresschange">the relative stress change between iterations must be below 10^(-stresschange). If the parameter is 0 it is ignored.</param>
        /// <returns>the resulting positions in the Euclidean space.
        /// Size: D x N, where D is the number of dimensions and N is the number of elements in the distance matrix (i.e. number of rows or columns)</returns>
        public static double[][] Mds(out string report, double[][] D, int dimensions, int rounds = 0, int timeout = 0, int stresschange = 0)
        {
            report = "";
            double[][] L = Data.landmarkMatrix(D);
            double[][] result = new double[dimensions][];
            for (int i=0;i<dimensions;i++) {
                result[i] = new double[D[0].Length];
            }

            ClassicalScaling.lmds(D, result);

            if ((rounds > 0) || (stresschange > 0) || (timeout > 0)) {
                StressMinimization sm = new StressMinimization(D, result, StressMinimization.weightMatrix(D, 0));

                report = sm.iterate(rounds, timeout, stresschange);
                double normalizedStress = sm.getNormalizedStress();
                report += ", normalized stress: " + normalizedStress;
            }

            return result;
        }
    }
}
