﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using ICSV.Interface.Operations;
using ICSV.Interface.Types;
using FlickrNet;
using System.Drawing;
using System.IO;
using System.Net;
using System.Diagnostics;
using System.Drawing.Imaging;
//using MathNet.Numerics.LinearAlgebra.Double;
using XnaGeometry;

using ICSV.Server.Signatures;
using ICSV.Server.FastEMD;
using ICSV.Server.Images;

namespace ICSV.Server
{
	/// <summary>
	/// Class responsible for chaining together the different stages of an image search.
	/// Chain:
	/// PrepareImages -> ImageProvider -> DistanceProvider -> CoordinateProvider -> PrepareImages
	/// </summary>
	class PrepareImages : CoordinateCallback
	{
		private static readonly int ALIVE_SIGNAL_TIMEOUT_SECONDS = 15; // AppHarbor default connection timeout = 30 seconds??
		private static readonly int MAX_ALLOWED_IMAGES = 1000;

		private static WebClient webClient = new WebClient();

		/// <summary>
		/// The request containing search query and image count for the current search.
		/// </summary>
		public readonly SearchImages Request;

		/// <summary>
		/// The server config.
		/// </summary>
		public readonly Interface.EMDConfig config;

		private readonly Object resultLock = new Object();

		private readonly HashSet<ImageContainer> processedImages = new HashSet<ImageContainer>();

		private readonly List<ImageData> resultList = new List<ImageData>();
		private readonly List<StatusData> statusList = new List<StatusData>();

		private AxisUpdate[] axes2D;
		private AxisUpdate[] axes3D;

		private CancellationTokenSource cancelToken;
		private bool processingFinished = false;

		private Stopwatch stopWatch = new Stopwatch();
		private Timer timer;

		/// <summary>
		/// Creates a new instance for the given search request and the given server config.
		/// </summary>
		/// <param name="request">the request</param>
		/// <param name="config">the configuration</param>
		public PrepareImages(SearchImages request, Interface.EMDConfig config)
		{
			this.Request = request;
			this.config = config;
		}

		/// <summary>
		/// Start a new search asynchronously for the search request (previously defined).
		/// This method must be called ONLY ONCE!
		/// </summary>
		public void Start()
		{
			stopWatch.Start();
			cancelToken = new CancellationTokenSource();
			CancellationToken token = cancelToken.Token;
			Task.Factory.StartNew(() => { DoComputation(token); }, token);
			startStillAliveTimer(ALIVE_SIGNAL_TIMEOUT_SECONDS); // send a still alive message every X seconds
		}

		/// <summary>
		/// Sends a cancellation signal to the whole processing chain.
		/// </summary>
		public void Cancel()
		{
			cancelToken.Cancel();
			AddStatus("Cancelled", StatusData.Type.CANCELLED);
			markAsFinished();
		}

		private void startStillAliveTimer(int timeoutSeconds)
		{
			long timeoutMilliseconds = timeoutSeconds * 1000;
			timer = new Timer(SendStillAliveSignal, null, timeoutMilliseconds, timeoutMilliseconds);
		}

		/// <summary>
		/// Send a still alive signal to the client (if it is listening) to avoid connection timeouts.
		/// </summary>
		/// <param name="state"></param>
		public void SendStillAliveSignal(object state)
		{
			lock (resultLock)
			{
				if (processingFinished)
				{
					timer.Change(Timeout.Infinite, Timeout.Infinite); // stop timer
					return;
				}
			}
			// add status + add fake result
			AddStatus("Server is still alive", StatusData.Type.STILL_ALIVE);
			AddResult(new ImageData() { Id = null, Url = null, pos2D = null, pos3D = null });
		}

		/// <summary>
		/// Returns all current status objects if at least 'minCount' objects are available.
		/// Otherwise this method blocks until the count is available or the search has finished.
		/// If the search has finished all status objects are returned, regardless of 'minCount'.
		/// </summary>
		/// <param name="minCount">the minimal count of status objects necessary for this method to succeed</param>
		/// <returns>a list of status objects</returns>
		public StatusData[] GetStatus(int minCount)
		{
			lock (statusList)
			{
				while (true)
				{
					bool returnResult = statusList.Count >= minCount || this.processingFinished;

					if (returnResult)
					{
						var status = statusList.ToArray();
						statusList.Clear();
						return status;
					}
					// wait for more results:
					Monitor.Wait(statusList);
				}
			}
		}

		/// <summary>
		/// Returns new results if available. This method blocks if no results are available.
		/// A maximum count of 50 results will be returned blockwise in order to keep the data traffic low for one request.
		/// </summary>
		/// <param name="finished">returns true if the search has finished and all remaining results are included in the current request</param>
		/// <param name="images">the image results. Contain data and locations.</param>
		/// <param name="axes2D">the 2D axes for this result</param>
		/// <param name="axes3D">the 3D axes for this result</param>
		public void GetResults(out bool finished, out ImageData[] images, out AxisUpdate[] axes2D, out AxisUpdate[] axes3D)
		{
			lock (resultLock)
			{
				while (true)
				{
					int maxElements = 50; // return small result 'packages' - don't stress the lousy server connection too much
					bool returnResult = resultList.Count > 0 || this.processingFinished;

					if (returnResult)
					{
						int objectCount = Math.Min(resultList.Count, maxElements);
						images = resultList.GetRange(0, objectCount).ToArray();
						resultList.RemoveRange(0, objectCount);

						finished = this.processingFinished && resultList.Count == 0;

						axes2D = this.axes2D == null ? new AxisUpdate[0] : this.axes2D;
						axes3D = this.axes3D == null ? new AxisUpdate[0] : this.axes3D;
						return;
					}
					// wait for more results:
					Monitor.Wait(resultLock);
				}
			}
		}

		/// <summary>
		/// Will be called by the CoordinatesProvider.
		/// This will trigger a rebuild of the axes and the results will be made available for the client.
		/// </summary>
		/// <param name="items">all images with coordinates so far</param>
		public void coordinatesUpdated(Dictionary<ImageContainer, CoordinateContainer> items)
		{
			// will be called by the CoordinatesProvider
			ImageContainer[] axisImages = getAxisImages();
			CoordinateContainer[] axisCoords = new CoordinateContainer[axisImages.Length];
			for (int i = 0; i < axisImages.Length; i++)
			{
				axisCoords[i] = items[axisImages[i]];
				items.Remove(axisImages[i]); // remove axes from subsequent calculations
			}

			AxesBuilder axesBuilder = new AxesBuilder();
			AxisUpdate[] axes2D = axesBuilder.CreateAxes(axisImages, axisCoords, items, 2);
			AxisUpdate[] axes3D = axesBuilder.CreateAxes(axisImages, axisCoords, items, 3);

			// publish results:
			lock (resultLock)
			{
				// update axes:
				this.axes2D = axes2D;
				this.axes3D = axes3D;

				// publish new images:
				foreach (var item in items)
				{
					var image = item.Key;
					var coord = item.Value;

					ImageData data = new ImageData();
					data.Id = image.Identifier;
					data.pos2D = convert(coord, 2);
					data.pos3D = convert(coord, 3);

					if (!processedImages.Contains(image))
					{
						// we never processed this image before
						// -> therefore send out the data
						data.Url = image.DataURI;
						processedImages.Add(image);
					}

					AddResult(data);
				}

				Monitor.PulseAll(resultLock);
			}
		}

		private void AddResult(ImageData result)
		{
			lock (resultLock)
			{
				resultList.Add(result);
			}
		}

		private static Interface.Types.ImagePosition convert(CoordinateContainer coord, int dimensions)
		{
			var c = coord.GetCoord(dimensions);
			var pos = new Interface.Types.ImagePosition();
			pos.XCoord = c.X;
			pos.YCoord = c.Y;
			pos.ZCoord = c.Z;
			return pos;
		}

		public void finished()
		{
			markAsFinished();
		}

		public void erroneous(String message)
		{
			AddStatus(message, StatusData.Type.FINISHED);
			markAsFinished();
		}

		public void info(String message)
		{
			AddStatus(message, StatusData.Type.INFO);
		}

		private ImageContainer[] getAxisImages()
		{
			var Lmin = ImageProvider.Instance.getLocalImage(config.WebHostImgPath + "Lmin.png");
			var Lmax = ImageProvider.Instance.getLocalImage(config.WebHostImgPath + "Lmax.png");
			var Amin = ImageProvider.Instance.getLocalImage(config.WebHostImgPath + "Amin.png");
			var Amax = ImageProvider.Instance.getLocalImage(config.WebHostImgPath + "Amax.png");
			var Bmin = ImageProvider.Instance.getLocalImage(config.WebHostImgPath + "Bmin.png");
			var Bmax = ImageProvider.Instance.getLocalImage(config.WebHostImgPath + "Bmax.png");
			return new ImageContainer[] { Lmin, Lmax, Amin, Amax, Bmin, Bmax };
		}

		private void DoComputation(CancellationToken token)
		{
			AddStatus("Beginning computation...", StatusData.Type.INFO);

			int photoCount = Math.Max(5, Math.Min(MAX_ALLOWED_IMAGES, Request.PhotoCount)); // A minimum of 5 should be enforced, otherwise MDS fails. A maximum is enforced as well

			Flickr flickr = new Flickr(config.FlickrKey, config.FlickrSecret);
			PhotoSearchOptions options = new PhotoSearchOptions();
			options.Tags = Request.ForQuery; // tags
			options.SortOrder = PhotoSearchSortOrder.Relevance;
			options.Accuracy = GeoAccuracy.World;

			// CHAIN: ImageProvider -> DistanceProvider -> CoordinateProvider -> this class
			CoordinateProvider coordinator = new CoordinateProvider(this, cancelToken.Token);
			DistanceProvider distancer = new DistanceProvider(coordinator, cancelToken.Token);
			coordinator.setDistanceProvider(distancer);

			// add images of the axes to the distance measurement, in order to find the positions of the axes later on
			var axisImages = getAxisImages();
			foreach (var axisImg in axisImages)
			{
				distancer.addImage(axisImg, true);
			}

			ImageProvider.Instance.searchImages(flickr, options, photoCount, distancer, cancelToken.Token);
		}

		/*
		 * Not necessary anymore - but keep this method just in case
		 */ 
		private static string ConvertToCorsUrl(string url)
		{
			// load images via CORS proxy -> therefore remove protocol and www:
			// see http://en.wikipedia.org/wiki/Cross-origin_resource_sharing

			System.Uri uri = new Uri(url);
			return "http://www.corsproxy.com/" + uri.Authority + uri.PathAndQuery;
		}

		private void markAsFinished()
		{
			lock (resultLock)
			{
				if (processingFinished) return; // already finished

				stopWatch.Stop();
				double time = Math.Round(stopWatch.ElapsedMilliseconds / 1000.0);
				AddStatus("Computation done in " + time + " seconds", StatusData.Type.FINISHED);

				processingFinished = true;
				Monitor.PulseAll(resultLock);
			}
		}

		private void AddStatus(String message, StatusData.Type type)
		{
			lock (statusList)
			{
				StatusData data = new StatusData();
				data.Id = (int) type;
				data.Message = message;
				statusList.Add(data);
				Monitor.PulseAll(statusList);
			}
		}
	}
}
