﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ICSV.Server.Signatures
{
    /// <summary>
    /// Helper class for kd clustering
    /// </summary>
    struct Element
    {
        public double[] coord;
        public double count;
    }

    /// <summary>
    /// Class responsible for creating signatures.
    /// </summary>
    class SignatureProvider
    {
        /// <summary>
        /// Creates a signature from the given parameters.
        /// 
        /// </summary>
        /// <param name="input">the CIE Lab converted pixels of an image. Assumed to be N x 3, where N is the number of pixels of the image and 3 are the axes of CIE Lab (L,a,b)</param>
        /// <param name="cellSize">The minimum cell size where subdivision stops. A smaller cell size typically results in more features of the signature</param>
        /// <param name="threshold">All clusters with a relative pixelCount (= pixelCountCluster / pixelCountImage) below this threshold are discarded.
        /// E.g. a threshold of 0.001 results in discarding clusters with a relative pixelCount of 0.1 % or less to be discarded.
        /// A higher threshold results in a smaller feature count, but delivers non-optimal results.
        /// </param>
        /// <returns>the created signature with a specific number of features (= clusters) and its corresponding count of pixel per feature</returns>
        public Signature CreateSignature(double[][] input, double cellSize, double threshold)
        {
            int elementCount = input.Length;
            List<Element> elements = new List<Element>(input.Length);

            for (int i = 0; i < elementCount; i++)
            {
                elements.Add(new Element() { coord = input[i], count = 1 });
            }

            // 1. pass: Create clusters from CIE Lab coordinates:
            List<Element> centroids = new List<Element>(50);
            Cluster(elements.ToArray(), centroids, 0, cellSize, 0);

            // 2. pass: Create clusters from centroids and shift the space by half the cell size:
            List<Element> result = new List<Element>(centroids.Count);
            Cluster(centroids.ToArray(), result, 0, cellSize, cellSize / 2);

            // 3. pass: Do a thresholding to remove clusters with not enough elements
            List<Feature> features = new List<Feature>(result.Count);
            List<double> weights = new List<double>(result.Count);

            double sumWeights = 0;
            for (int i = 0; i < result.Count; i++)
            {
                Element centroid = result[i];
                double relativeCount = centroid.count / elementCount;

                if (relativeCount >= threshold) // we don't accept centroid below a certain count
                {
                    features.Add(new Feature3D(centroid.coord[0], centroid.coord[1], centroid.coord[2]));
                    weights.Add(centroid.count);
                    sumWeights += centroid.count;
                }
            }

            // 4. pass: Normalize weights according to centroids which are really part of the signature and create signature
            for (int i = 0; i < weights.Count; i++)
            {
                weights[i] *= elementCount / sumWeights; // weights should sum up to 'elementCount'
            }
            Signature signature = new Signature();
            signature.setFeatures(features.ToArray());
            signature.setNumberOfFeatures(features.Count);
            signature.setWeights(weights.ToArray());

            return signature;
        }

        private void Cluster(Element[] input, List<Element> output, int dimension, double cellSize, double splitOffset)
        {
            if (input.Length == 0) return; // nothing to split anymore

            // find min/max for the current dimension:
            double min = Double.MaxValue;
            double max = double.MinValue;
            for (int i = 0; i < input.Length; i++)
            {
                min = Math.Min(min, input[i].coord[dimension]);
                max = Math.Max(max, input[i].coord[dimension]);
            }

            double split = (max - min) / 2;
            if (split >= cellSize)
            {
                // split
                double splitPos = min + split + splitOffset;

                List<Element> left = new List<Element>(input.Length / 2);
                List<Element> right = new List<Element>(input.Length / 2);
                for (int i = 0; i < input.Length; i++)
                {
                    double coordinate = input[i].coord[dimension];
                    if (coordinate <= splitPos)
                    {
                        left.Add(input[i]);
                    }
                    else
                    {
                        right.Add(input[i]);
                    }
                }

                int nextDimension = (dimension + 1) % 3;
                Cluster(left.ToArray(), output, nextDimension, cellSize, splitOffset); // left branch
                Cluster(right.ToArray(), output, nextDimension, cellSize, splitOffset); // right branch
            }
            else
            {
                // create new cluster
                Element cluster = new Element();
                cluster.coord = new double[3] { 0, 0, 0 };
                cluster.count = 0;

                // average position and count elements:
                for (int i = 0; i < input.Length; i++)
                {
                    Element element = input[i];

                    for (int j = 0; j < 3; j++)
                    {
                        cluster.coord[j] += element.coord[j];
                    }
                    cluster.count += element.count;
                }

                // averaging:
                cluster.coord[0] /= input.Length;
                cluster.coord[1] /= input.Length;
                cluster.coord[2] /= input.Length;

                output.Add(cluster);
            }
        }
    }
}
